/*
 Frequ systh forTracker HF radio.
 VK3ZYZ 20211208
 NOTE. The encoder needs to be changed to interupt driven.
 */

#define Version "Tracker20211208"

#include "si5351.h"
#include "Wire.h"
//#include <LCD.h>  
#include <LiquidCrystal_I2C.h>

// synthesiser 3.5Mhz-3.9MHz (80m band)
unsigned long frequency =   3630000;         // 4 bytes required for frequency
unsigned long step_size =      1000;         // default tuning increment (Hz)
unsigned long offset    =   1647000;         // IF offset for clockgen
unsigned long USB       = 165000000;         // USB sideband mixer crystal
unsigned long LSB       = 164700000;         // LSB sideband mixer crystal                         
/*
unsigned long frequency0 =    00000;          // 4 bytes required for frequency
unsigned long frequency1 =  1810000;          // 4 bytes required for frequency
unsigned long frequency2 =  3525000;          // 4 bytes required for frequency 
unsigned long frequency3 =  5351000;          // 4 bytes required for frequency 
unsigned long frequency4 =  7000000;          // 4 bytes required for frequency
unsigned long frequency5 = 10100000;          // 4 bytes required for frequency
unsigned long frequency6 = 14100000;          // 4 bytes required for frequency
unsigned long frequency7 = 18068000;          // 4 bytes required for frequency
unsigned long frequency8 = 21000000;          // 4 bytes required for frequency
*/
unsigned long frequency0 =    00000;          // 4 bytes required for frequency
unsigned long frequency1 =  1810000;          // 4 bytes required for frequency
unsigned long frequency2 =  3525000;          // 4 bytes required for frequency 
unsigned long frequency3 =  3625000;          // 4 bytes required for frequency 
unsigned long frequency4 =  7000000;          // 4 bytes required for frequency
unsigned long frequency5 = 10120000;          // 4 bytes required for frequency
unsigned long frequency6 = 14100000;          // 4 bytes required for frequency
unsigned long frequency7 =    00000;          // 4 bytes required for frequency
unsigned long frequency8 =    00000;          // 4 bytes required for frequency

int temp = 0;  
static signed char ham_band_index = 0;
int32_t cal_factor = 27000; //  eg. 16360 for my board, found from Si5351 calibrator sketch.
  
long encoder;                                // encoder direction: CCW=-1; NC=0; CW=1
Si5351 clockgen;                             // instantiate clockgen


// encoder definitions
#define ENCODER_A      2
#define ENCODER_B      3
#define ENCODER_BUTTON 4

#define PTT            6
//Band Select Outputs
#define BandF1         7
#define BandF2         8
#define BandF3         9
#define BandF4        10

#define SSB_SWITCH    A2
#define Clarifier     A3
// A4 = SDA
// A5 = SCL
#define BAND_SWITCH   A6

                          
#define ENCODER_INPUT_PORT PIND
int sideband = 0;   
int last_sideband = 1;

LiquidCrystal_I2C  lcd(0x27,16,2); 


void setup() { 
  Serial.begin(115200);                 // set serial port speed
  lcd.init();                            // 16 char x 2 line LCD display
  lcd.backlight();
  lcd.print(Version);
  delay(1000);
  lcd.clear();
  
  bool i2c_found;

  // Start serial and initialize the Si5351
  Serial.begin(57600);
  i2c_found = clockgen.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);
  clockgen.set_correction(cal_factor, SI5351_PLL_INPUT_XO); // <<<<<<<this is found from Si5351 calibrator code.

  if(!i2c_found)
  {
    lcd.print("Device not found on I2C bus!");
  }
  
  // configure rotary encoder & push-button
  pinMode(ENCODER_A, INPUT_PULLUP);          // WIRE-OR encoder leads
  pinMode(ENCODER_B, INPUT_PULLUP);  
  pinMode(ENCODER_BUTTON, INPUT_PULLUP);
  pinMode(BAND_SWITCH, INPUT_PULLUP);
  pinMode(SSB_SWITCH, INPUT); // High = USB, LOW =  LSB

  pinMode(BandF1,OUTPUT); // 2.000 -  3.300Mhz
  pinMode(BandF2,OUTPUT); // 3.301 -  5.500Mhz
  pinMode(BandF3,OUTPUT); // 5.501 -  9.000Mhz
  pinMode(BandF4,OUTPUT); // 9.001 - 15.000Mhz

/*
       HF Bands              SSB part                      LPF band
 160M =  1.800 -  1.875 (1.840-1.875)                     ` 1 maybe ??????
  80M =  3.500 -  3.800 (3.535-3.620, 3.640-3.700)          2
  60M =  5.3515 – 5.3665 MHz
  40M =  7.000 -  7.300 ( 7.040-7.300)                      3
  20M = 14.000 - 14.350 (14.112-14.350)                     4
  
  17M = 18.068 - 18.168 (18.110-18.168)                     
  15M = 21.000 - 21.240 (21.150-21.450)                     
  12M = 24.890 - 24.990 (24.930-24.990)                     
  10M = 28.000 - 29.700 (28.300-29.100)   
  
 */
 // Active LOW to select BandFilter
  digitalWrite(BandF1,LOW);
  digitalWrite(BandF2,HIGH);
  digitalWrite(BandF3,HIGH);
  digitalWrite(BandF4,HIGH);
  delay(200);
  digitalWrite(BandF1,HIGH);
  digitalWrite(BandF2,LOW);
  digitalWrite(BandF3,HIGH);
  digitalWrite(BandF4,HIGH);
  delay(200);
  digitalWrite(BandF1,HIGH);
  digitalWrite(BandF2,HIGH);
  digitalWrite(BandF3,LOW);
  digitalWrite(BandF4,HIGH);
  delay(200);
  digitalWrite(BandF1,HIGH);
  digitalWrite(BandF2,HIGH);
  digitalWrite(BandF3,HIGH);
  digitalWrite(BandF4,LOW);
  delay(200);

  set_synth();
  display_freq_step();               // display frequency & step-size

}

/******************************
 main loop
*******************************/
void loop() {
  check_band();
  check_step_size();
 
  if (digitalRead(SSB_SWITCH)!=last_sideband) { 
    set_synth();
    display_freq_step();                          // display frequency & step-size
  }
 
  // check the frequency 
  encoder = read_encoder();                       // returns +1/0/-1 (cw/nc/ccw)
  if (encoder != 0) {
    frequency = frequency + (encoder * step_size);
    
    set_synth();
    display_freq_step();                          // display frequency & step-size
  } 
}

/******************************
 read encoder direction
*******************************/
char read_encoder() {
  // definitions
  #define DELAY 100                               // number of polling-loops   
  static unsigned short integrator;               // encoder debounce
 
  /* the array values are determined by combining the previous and current A1,A0 
  Gray Code bit-patterns to form an array-index. +1 is assigned to all clockwise 
  (CW) indexes; -1 is assigned to all counter-clockwise (CCW) indexes; and 0 is 
  assigned to the remaining positions to indicate no movement*/   
  static char encoder[] = {0,-1,1,0,1,0,0,-1,-1,0,0,1,0,1,-1,0}; // SWAPPED PHASES
  //   static char encoder[] = {0,1,-1,0,-1,0,0,1,1,0,0,-1,0,-1,1,0};
  static unsigned char encoder_index = 0;
  
  static long count = 0;                         // CW rotation = +1; CCW rotation = -1
  static long this_reading;
  static long last_reading;
  char value;                                    // return value

  /* check for encoder rotation. The absolute "count" value changes as the encoder rotates.
  Typically there are four states (counts) per click, but often more due to contact bounce.
  Bounce doesn't matter so long as the "count" decreases" for CCW rotation; remains constant 
  when there is no rotation; and increases with CW rotation. Bounce can be completely 
  eliminated if we look for the "detent" code and integrate as shown below. */  
  encoder_index >>= 2;                           // shift previous bit-pattern right two places
  encoder_index |= (ENCODER_INPUT_PORT & 0x0C);  // get current bit-pattern and combine with previous
  count += encoder[( encoder_index & 0x0f )];     
  
  // debounce encoder
  if ((ENCODER_INPUT_PORT & 0x0C) == 0x0C) {        // encoder on an indent (3 decimal == 00000011 binary)
    if (integrator < DELAY) {                    
      integrator++;
    } 
  } else {                                       // encoder not resting on indent
     if (integrator > 0) {
       integrator--;
     }
  }
  
  // prepare the "return" value
  if (integrator >= DELAY) {                     // encoder deemed to be stationary
    this_reading = count;
    if (this_reading < last_reading) value = -1;
    if (this_reading == last_reading) value = 0;
    if (this_reading > last_reading) value = 1;
    last_reading = this_reading;                 // both readings now hold current "count"
    integrator = 0;                              // reset integrator
  } else {
    value = 0;
  }
  return(value);
}
  
/******************************
 check frequency band
*******************************/
void check_band() {               // band now set from A6 reading

  temp = (analogRead(A6));
while ( temp !=(analogRead(A6)))
  {  temp = (analogRead(A6));}      // read until 2 the same

if (temp <= 40)
  { temp = 0; 
  } 
  
if (temp >= 600)    
  { temp = 8; 
  }
  
if (temp >= 400)  
  { temp = 7; 
  }

if (temp >= 300)   
  { temp = 6; 
  }  

if (temp >= 160)    
  { temp = 5; 
  }

if (temp >= 100)    
  { temp = 4; 
  }

if (temp >= 80)    
  { temp = 3; 
  }

if (temp >= 60)    
  { temp = 2; 
  }
  
if (temp >= 50)     
  { temp = 1; 
  }



   
if (ham_band_index!=temp){    // update the band if it has changed
      ham_band_index=temp;
      
  long ham_band[] = {frequency0,frequency1,frequency2,frequency3,frequency4,frequency5,frequency6,frequency7,frequency8};
        frequency = ham_band[ham_band_index];     // change frequency band  
        set_synth ();                             // set the new frequ ****** added 20190502
        display_freq_step(); 
}
}

/******************************
 check the step size
*******************************/
void check_step_size() {
  // delay definitions
  #define DELAY 3000                             // DELAY = debounce(mS)*loops/second
  #define LONG_PUSH 15000                        // number of loops considered "long"
  
  // variables
  static boolean transition = false;             // valid during button push
  static unsigned short integrator;              // varies from 0 to MAXIMUM
  static unsigned short push_duration;           // determines step size/direction
  
  // look for first high-to-low transition
  if (!transition && (digitalRead(ENCODER_BUTTON) == LOW)) {
    transition = true;
    integrator = 0;
    push_duration = 0;
  }
  
  // look for last low-to-high transition
  if (transition) {
    if (push_duration < (LONG_PUSH*2)) {         // track push duration
      push_duration++;
    }
    if (digitalRead(ENCODER_BUTTON) == HIGH) {   // ENCODER_BUTTON is HIGH
      if (integrator < DELAY) {        
        integrator++;
      } else {                                   // ENCODER_BUTTON is LOW
        if (integrator > 0) {
    integrator--;
  }
      }
    }
   
    // update the step_size
    if (integrator >= DELAY) { 
      transition = false;
      if (push_duration > LONG_PUSH) {          // LONG PUSH: decrease step size
        step_size = step_size/10;
        if (step_size < 10) {
          step_size = 100000;                   // wrap around    
        }        
      } else {                                  // SHORT PUSH: increase step size
        step_size = step_size*10;
        if (step_size > 100000) {
          step_size = 10;                       // wrap around
        }         
      }
     
      display_freq_step();                            // update step-size     
    }
  }
}


/*******************************
set the Si5351 CLK0 frequency
 ******************************/

 void set_synth ()
{
  switch (ham_band_index)
   {       // limit to SSB part of bands  
      
    case 0:
    frequency = constrain(frequency, 0000000, 00000);   // out of band
    frequency0 = frequency;
    digitalWrite(BandF1,HIGH);
    digitalWrite(BandF2,HIGH);
    digitalWrite(BandF3,HIGH);
    digitalWrite(BandF4,HIGH);
    break;
  
    case 1:
    frequency = constrain(frequency, 1800000, 1875000);
    frequency1 = frequency;
    digitalWrite(BandF1,LOW);
    digitalWrite(BandF2,HIGH);
    digitalWrite(BandF3,HIGH);
    digitalWrite(BandF4,HIGH);
    break;
  
    case 2:
    frequency = constrain(frequency, 3500000, 3670000); 
    frequency2 = frequency;
    digitalWrite(BandF1,HIGH);
    digitalWrite(BandF2,LOW);
    digitalWrite(BandF3,HIGH);
    digitalWrite(BandF4,HIGH);
    break;

    case 3:
    frequency = constrain(frequency, 3600000, 3700000);
    frequency3 = frequency;
    digitalWrite(BandF1,HIGH);
    digitalWrite(BandF2,LOW);
    digitalWrite(BandF3,HIGH);
    digitalWrite(BandF4,HIGH);
    break;

    case 4:
    frequency = constrain(frequency, 7000000, 7300000);
    frequency4 = frequency;
    digitalWrite(BandF1,HIGH);
    digitalWrite(BandF2,HIGH);
    digitalWrite(BandF3,LOW);
    digitalWrite(BandF4,HIGH);
    break;
  
    case 5: 
    frequency = constrain(frequency, 10100000, 10150000);
    frequency7 = frequency;
    digitalWrite(BandF1,HIGH);
    digitalWrite(BandF2,HIGH);
    digitalWrite(BandF3,HIGH);
    digitalWrite(BandF4,LOW);
    break;

    case 6: 
    frequency = constrain(frequency, 14112000, 14350000);
    frequency8 = frequency;
    digitalWrite(BandF1,HIGH);
    digitalWrite(BandF2,HIGH);
    digitalWrite(BandF3,HIGH);
    digitalWrite(BandF4,LOW);
    break;

    case 7:
    frequency = constrain(frequency, 0000000, 00000);   // out of band
    frequency0 = frequency;
    digitalWrite(BandF1,HIGH);
    digitalWrite(BandF2,HIGH);
    digitalWrite(BandF3,HIGH);
    digitalWrite(BandF4,HIGH);
    break;

    case 8:
    frequency = constrain(frequency, 0000000, 00000);   // out of band
    frequency0 = frequency;
    digitalWrite(BandF1,HIGH);
    digitalWrite(BandF2,HIGH);
    digitalWrite(BandF3,HIGH);
    digitalWrite(BandF4,HIGH);
    break;
    
   }

 if (digitalRead(SSB_SWITCH)==HIGH)
    {
    clockgen.set_freq(USB, SI5351_CLK0); // USB
    last_sideband=1;
    }
  else
    {
    clockgen.set_freq(LSB, SI5351_CLK0); // LSB
    last_sideband=0;
    }

clockgen.set_freq(((frequency+offset)*100), SI5351_CLK1);
}

/********************************
 update frequency and step-size
********************************/
void display_freq_step() {
  long MHz = frequency/1000000;
   
  // convert frequency to floating point
  // adding 0.0000001 fixes a rounding error
  double freq = (double) frequency/1000000+0.0000001;

  // display frequency
  lcd.clear();
      if (digitalRead(SSB_SWITCH)==HIGH)
      { lcd.print("U");
      } 
      else 
      { lcd.print("L");
      }    

      // DEBUG - display ADC read od band switch     
     lcd.setCursor(11,1);
     lcd.print(analogRead(A6));
     lcd.setCursor(15,1);   
     lcd.print(temp);        
  
  if (MHz < 10) {
    lcd.setCursor(3,0);                          // start at char 1, line 1
  } else {
    lcd.setCursor(2,0);                          // start at char 1, line 1
  }
    lcd.print(freq, 6);
    lcd.print(" MHz ");
  
  // last digit always zero                      // digit sometimes showed a 1
  lcd.setCursor(10,0);
  lcd.print("0");

  // display step size
  switch (step_size) {
    case 10:
      lcd.setCursor(9,1);
      lcd.print("*");       
      break;
    case 100:
      lcd.setCursor(8,1);
      lcd.print("*");       
      break;
    case 1000:
      lcd.setCursor(7,1);
      lcd.print("*");       
      break;
    case 10000:
      lcd.setCursor(6,1);
      lcd.print("*");       
      break;
    case 100000:
      lcd.setCursor(5,1);
      lcd.print("*");       
      break;
    default:
      break;
  }

  // display the band in meters
      lcd.setCursor(0,1);
  switch (ham_band_index)
   {  case 1:
      lcd.print("160M");
      break;
      case 2:
      lcd.print("80MA");
      break;
      case 3:
      lcd.print("80MB");
      break;
      case 4:
      lcd.print(" 40M");
      break;
      case 5:
      lcd.print(" 30M");
      break;
      case 6:
      lcd.print(" 20M");
      break;
   }  
}
